-- GPU

local vs = [[
    #version 300 es
    in vec2 position;
    void main() {
        gl_Position = vec4(position, 0.0, 1.0);
    }
]]

local fs = [[
    #version 300 es
    precision highp float;

    uniform int xsSize;
    uniform highp sampler2D xs;
    uniform highp sampler2D ws;

    out vec4 outValue;

    vec4 packF16(float f)
    {
        uint packed = packHalf2x16(vec2(f, 0.0));
        vec4 o = vec4((packed>>8)&uint(0xff), packed&uint(0xff), 0.0, 255.0);
        return o / 255.0;
    }

    float unpackF16x1(vec4 packed)
    {
        uvec4 upacked = uvec4(round(packed * 255.0));
        vec2 f = unpackHalf2x16((upacked.x<<8) + upacked.y);
        return f.x;
    }

    vec2 unpackF16x2(vec4 packed)
    {
        uvec4 upacked = uvec4(round(packed * 255.0));
        float f1 = unpackHalf2x16((upacked.x<<8) + upacked.y).x;
        float f2 = unpackHalf2x16((upacked.z<<8) + upacked.w).x;
        return vec2(f1, f2);
    }

    // Sample 2 floats from a texture's texel
    vec2 sampleF16x2(sampler2D s, int x, int y)
    {
        return unpackF16x2(texelFetch(s, ivec2(x, y), 0));
    }

    // Sample 1 float from a texture's texel
    float sampleF16x1(sampler2D s, int x, int y)
    {
        return unpackF16x1(texelFetch(s, ivec2(x, y), 0));
    }
    
    void main()
    {
        highp float act = sampleF16x1(ws, int(gl_FragCoord.y), 0);

        for (int i = 0, j = 1; i < xsSize; i+=2, j++)
        {
            // Read 2 input values
            float i1 = sampleF16x1(xs, 0, i);
            float i2 = sampleF16x1(xs, 0, i+1);

            // Read corresponding weights
            vec2 w = sampleF16x2(ws, int(gl_FragCoord.y), j);

            // Multiply
            act += dot(vec2(i1, i2), w);
        }

        // Normalise
        act = tanh(act);

        // Encode & output
        outValue = packF16(act);
    }
]]

local sh, m
local function initGPU()
    sh = shader(vs, fs)
    m = mesh()
    m.vertices = {
        vec2(-1, -3),
        vec2(-1, 1),
        vec2(3, 1)
    }
    m.shader = sh
end

local function unpackF16x1(v)
    local exp = ((v.x>>2)&0x1F) - 15
    local impliedBit = 0
    if exp == -15 then -- special case of IEEE-754
        exp = -14
    else
        impliedBit = 1
    end
    local sign = ((v.x >> 7 > 0) and -1) or 1
    local mantissa = ((v.x&0x3)*256) + v.y
    --print(sign, exp, impliedBit, mantissa)
    return sign * (2^exp * (impliedBit + mantissa/1024))
end
local function packF16x1(v)
    local binaryv = string.unpack("I4", string.pack("f", v))
    local exp = (((binaryv >> 23) & 0xFF) - 0x7F) + 0xF                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       
    local pk = ((binaryv >> 16) & 0x8000) -- sign bit
    pk = pk | ((exp << 10) & 0x7C00) -- exp
    pk = pk | ((binaryv >> 13) & 0x03FF) -- mantissa
    pk = vec4(pk >> 8, pk & 0xFF, 0, 255)
    return pk
end
local function packF16x2(v1, v2)
    pk1 = packF16x1(v1)
    pk2 = packF16x1(v2)
    return vec4(pk1.x, pk1.y, pk2.x, pk2.y)
end

-- Non-retina scaled image constructor
local _image = image
local function image(...)
    local img = _image(...)
    saveImage(asset .. "tmp.png", img)
    saveImage(asset .. "tmp@2x.png", nil)
    img = readImage(asset .. "tmp.png")
    saveImage(asset .. "tmp@2x.png", nil)
    return img
end

local GPUIOImage = class()
function GPUIOImage:init(size)
    self.img = image(1, size + size%2)
    self.size = size
end

function GPUIOImage:set(t)
    for i,v in ipairs(t) do
        v = packF16x1(v)
        self.img:set(1, i, v.x, v.y, v.z, v.w)
    end
    sprite(self.img, 0, 0, 0, 0) -- Update GPU textureObject
end

function GPUIOImage:get()
    local t = {}
    for i=1,self.size do
        table.insert(t, unpackF16x1(vec4(self.img:get(1,i))))
    end
    return t
end

local GPULayer = class()
function GPULayer:init(nin, nout)
    self.nin = nin
    self.nout = nout
    self.numParams = nout*(1+nin)
    
    -- +1 for bias
    local paramTw = nout
    local paramTh = math.ceil(nin / 2) + 1
    
    self.param = image(paramTw, paramTh)
    self.out = GPUIOImage(nout)
    
    --print(string.format("Layer with %dx inputs, %dx outputs, %dx params", nin, nout, self.numParams))
end

function GPULayer:setParameters(params)
    -- For each neuron
    for x = 1, self.param.width do
        
        -- Index base
        local base = ((x-1) * (self.nin+1)) + 1
        
        -- Set bias pixel
        local i = x * (self.nin+1)
        local v = packF16x1(params[i])
        self.param:set(x, 1, v.x, v.y, 0, 0)
        
        -- Set weight pixels
        for y = 2, self.param.height do
            local i = base + ((y-2)*2)
            
            -- Pack next 2 weights
            local v = packF16x2(params[i], params[i+1])
            self.param:set(x, y, v.x, v.y, v.z, v.w)
        end
    end
    
    sprite(self.param, 0, 0, 0, 0) -- Update GPU textureObject
end

function GPULayer:randomiseParameters()
    local img = self.param
    
    -- Randomise biases in first row
    for x = 1,img.width do
        local v = encodeF16x2(math.random()*2 - 1, 0)
        img:set(x, 1, v.x, v.y, v.z, v.w);
    end
    
    -- Randomise weights in other rows
    for x = 1,img.width do
        for y = 2,img.height do
            local v = encodeF16x2(math.random()*2 - 1, math.random()*2 - 1)
            img:set(x, y, v.x, v.y, v.z, v.w);
        end
    end
    sprite(img, 0, 0, 0, 0) -- Update GPU textureObject
end

function GPULayer:__call(input)
    assert(input ~= nil, "Layer input must be provided!")
    
    noSmooth()
    setContext(self.out.img)
    background(0, 0)
    
    m.shader.ws = self.param
    m.shader.xs = input.img
    m.shader.xsSize = self.nin
    m:draw()
    
    setContext()
    
    -- Return output image
    return self.out
end

local GPUMLP = class()
function GPUMLP:init(nin, nouts)
    self.numParams = 0
    
    -- Init layers
    table.insert(nouts, 1, nin)
    self.layers = {}
    for i = 1,#nouts-1 do
        assert(nouts[i] % 2 == 0, "Layer sizes must be multiples of 2 to use GPU acceleration!")
        local l = GPULayer(nouts[i], nouts[i+1])
        table.insert(self.layers, l)
        self.numParams = self.numParams + (l.nout * (l.nin+1))
    end
    
    -- Init input buffer
    self.input = GPUIOImage(nin)
end

function GPUMLP:setParameters(params)
    local i = 1
    for _,layer in ipairs(self.layers) do
        local t = {}
        table.move(params, i, i + layer.numParams - 1, 1, t)
        layer:setParameters(t)
        i = i + layer.numParams
    end
end

function GPUMLP:randomiseParameters()
    for _,layer in ipairs(self.layers) do
        layer:randomiseParameters()
    end
end

function GPUMLP:__call(input)
    self.input:set(input)
    input = self.input
    for _,layer in ipairs(self.layers) do
        input = layer(input)
    end
    return input
end

local GPUModel = class()
function GPUModel.load(from)
    local jsonArr = readText(from)
    local arr = json.decode(jsonArr)
    return GPUModel(arr.nin, arr.nouts, arr.data)
end

function GPUModel:init(nin, nouts, params)
    self.nin = nin
    self.nouts = nouts
    self.mlp = GPUMLP(nin, nouts)
    
    -- Configure the model parameters
    if params then
        self:setParameters(params)
    end
    
    -- Initialise
    initGPU()
    
    print("GPU Model with " .. self.mlp.numParams .. " parameters")
end

function GPUModel:setParameters(params)
    self.mlp:setParameters(params)
end

function GPUModel:paramArray()
    local out = {}
    for i,p in ipairs(self.params) do
        out[i] = p.data
    end
    return out
end

function GPUModel:save(to)
    assert(false, "GPUModel:save() Unimplemented!")
end

function GPUModel:__call(input)
    return self.mlp(input):get()
end

AI.GPUModel = GPUModel