local Model = class()

function Model.load(from, useGPU)
    if useGPU == true then
        return AI.GPUModel.load(from)
    end
    
    local jsonArr = readText(from)
    local arr = json.decode(jsonArr)
    
    local m = Model(arr.nin, arr.nouts)
    for i,p in ipairs(arr.data) do
        m.params[i].data = p
    end
    
    return m
end

function Model:init(nin, nouts, params)
    self.nin = nin
    self.nouts = nouts
    
    self.mlp = AI.MLP(nin, nouts)
    self.params = self.mlp:parameters()
    
    -- Configure the model parameters
    if params then
        for i,p in ipairs(params) do
            self.params[i].data = p
        end
    end
        
    print("CPU Model with " .. #self.params .. " parameters")
end

function Model:paramArray()
    local out = {}
    for i,p in ipairs(self.params) do
        out[i] = p.data
    end
    return out
end

function Model:save(to)
    local arr = {}
    for i,p in ipairs(self.params) do
        arr[i] = p.data
    end
    arr = json.encode({
            nin = self.nin,
            nouts = self.nouts,
            data = arr
        }, {indent=true})
    saveText(to, arr)
    print("Saved " .. #self.params .. " parameters.")
end

function Model:train(dataset, rate, count)
    local loss
    
    local xs = dataset[1]
    local ys = dataset[2]
    
    -- Training loop
    for k = 1, count do
        
        -- Forward pass
        local ypred = {}
        for i,x in ipairs(xs) do
            ypred[i] = self.mlp(x)
        end
        loss = 0
        for i = 1,#ys do
            local ygt = ys[i]
            local yout = ypred[i]
            for j = 1,#ygt do
                loss = loss + (yout[j] - ygt[j])^2
            end
        end
        
        -- Backward pass
        for _,p in ipairs(self.params) do
            p.grad = 0
        end
        loss:backward()
        
        -- Update
        for _,p in ipairs(self.params) do
            p.data = p.data - (rate * p.grad)
        end
        
        print(k, loss.data)
        
        if loss.data ~= loss.data then
            error("loss explosion")
        end
    end
    
    return loss.data
end

function Model:__call(input)
    local out = {}
    for i,v in ipairs(self.mlp(input)) do
        out[i] = v.data
    end
    return out
end

AI.Model = Model